/*
 * Copyright 2010-13 Fraunhofer ISE
 *
 * This file is part of j62056.
 * For more information visit http://www.openmuc.org
 *
 * j62056 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * j62056 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with j62056.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.j62056;

import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class Connection {

	private final String serialPortName;
	private SerialPort serialPort;

	private final boolean handleEcho;
	private final boolean noBaudRateChange;

	private DataOutputStream os;
	private DataInputStream is;

	private static final byte[] REQUEST_MESSAGE = new byte[] { (byte) 0x2F, (byte) 0x3F, (byte) 0x21, (byte) 0x0D,
			(byte) 0x0A };

	private static final byte[] ACKNOWLEDGE = new byte[] { (byte) 0x06, (byte) 0x30, (byte) 0x30, (byte) 0x30,
			(byte) 0x0D, (byte) 0x0A };

	private static final int INPUT_BUFFER_LENGTH = 1024;
	private final byte[] buffer = new byte[INPUT_BUFFER_LENGTH];

	/**
	 * Creates a Connection object. You must call <code>open()</code> before calling <code>read()</code> in order to
	 * read data.
	 * 
	 * @param serialPort
	 *            examples for serial port identifiers on Linux are "/dev/ttyS0" or "/dev/ttyUSB0" and on Windows "COM1"
	 * @param handleEcho
	 *            tells the connection to throw away echos of outgoing messages. Echos are caused by some optical
	 *            transceivers.
	 * @param noBaudrateChange
	 *            tells the connection not change the baud rate during message exchange and stay at 300 instead.
	 *            Normally the connection will change the baud rate after an initial handshake with the remote device.
	 *            This may cause problems if the serial port does not change its baud rate quickly enough to receive the
	 *            full data message from the remote device. Therefore the baud rate change can be disabled using this
	 *            setting.
	 */
	public Connection(String serialPort, boolean handleEcho, boolean noBaudrateChange) {
		if (serialPort == null) {
			throw new IllegalArgumentException("serialPort may not be NULL");
		}

		serialPortName = serialPort;
		this.handleEcho = handleEcho;
		noBaudRateChange = noBaudrateChange;
	}

	/**
	 * Creates a Connection object. The options handleEcho and noBaudrateChange are set to false.
	 * 
	 * @param serialPort
	 *            examples for serial port identifiers on Linux are "/dev/ttyS0" or "/dev/ttyUSB0" and on Windows "COM1"
	 */
	public Connection(String serialPort) {
		this(serialPort, false, false);
	}

	/**
	 * Opens the serial port associated with this connection.
	 * 
	 * @throws IOException
	 *             if any kind of error occurs opening the serial port.
	 */
	public void open() throws IOException {

		CommPortIdentifier portIdentifier;
		try {
			portIdentifier = CommPortIdentifier.getPortIdentifier(serialPortName);
		} catch (NoSuchPortException e) {
			throw new IOException("Serial port with given name does not exist", e);
		}

		if (portIdentifier.isCurrentlyOwned()) {
			throw new IOException("Serial port is currently in use.");
		}

		CommPort commPort;
		try {
			commPort = portIdentifier.open(this.getClass().getName(), 2000);
		} catch (PortInUseException e) {
			throw new IOException("Serial port is currently in use.", e);
		}

		if (!(commPort instanceof SerialPort)) {
			commPort.close();
			throw new IOException("The specified CommPort is not a serial port");
		}

		serialPort = (SerialPort) commPort;

		try {
			os = new DataOutputStream(serialPort.getOutputStream());
			is = new DataInputStream(serialPort.getInputStream());
		} catch (IOException e) {
			serialPort.close();
			throw new IOException("Error getting input or output or input stream from serial port", e);
		}

	}

	/**
	 * Requests a data message from the remote device using IEC 62056-21 Mode C. The data message received is parsed and
	 * a list of data sets is returned.
	 * 
	 * @param timeout
	 *            the maximum time in milliseconds that the function shall wait for a response. A timeout of zero is
	 *            interpreted as an infinite timeout.
	 * @return A list of data sets contained in the data message response from the remote device or <code>null</code> if
	 *         the specified waiting time elapses before a data message was received.
	 * @throws IOException
	 *             if any kind of error occurs while trying to read the remote device. The connection is not closed when
	 *             an IOException is thrown.
	 */
	public List<DataSet> read(int timeout) throws IOException {

		try {
			serialPort.setSerialPortParams(300, SerialPort.DATABITS_7, SerialPort.STOPBITS_1, SerialPort.PARITY_EVEN);
		} catch (UnsupportedCommOperationException e) {
			close();
			throw new IOException("Unable to set the given serial comm parameters", e);
		}

		os.write(REQUEST_MESSAGE);
		os.flush();
		int baudRateSetting;
		int baudRate;

		boolean readSuccessful = false;
		int timeval = 0;
		int numBytesRead = 0;

		while (timeout == 0 || timeval < timeout) {
			if (is.available() > 0) {
				numBytesRead += is.read(buffer, numBytesRead, INPUT_BUFFER_LENGTH - numBytesRead);

				if (((handleEcho && numBytesRead > 11) || (!handleEcho && numBytesRead > 6))
						&& buffer[numBytesRead - 2] == 0x0D && buffer[numBytesRead - 1] == 0x0A) {
					readSuccessful = true;
					break;
				}
			}

			try {
				Thread.sleep(50);
			} catch (InterruptedException e) {
			}

			timeval += 50;
		}

		if (!readSuccessful) {
			return null;
		}

		int offset = 0;
		if (handleEcho) {
			offset = 5;
		}

		baudRateSetting = buffer[offset + 4];
		baudRate = getBaudRate(baudRateSetting);

		if (buffer[offset + 5] != 0x5C || buffer[offset + 6] != 0x32 || baudRate == -1) {
			close();
			throw new IOException("Syntax error in identification message received");
		}

		if (!noBaudRateChange) {
			ACKNOWLEDGE[2] = (byte) baudRateSetting;
		}
		os.write(ACKNOWLEDGE);
		os.flush();

		if (handleEcho) {
			readSuccessful = false;
			numBytesRead = 0;
			while (timeout == 0 || timeval < timeout) {
				if (is.available() > 0) {
					numBytesRead += is.read(buffer, 0, ACKNOWLEDGE.length - numBytesRead);
					if (numBytesRead == ACKNOWLEDGE.length) {
						readSuccessful = true;
						break;
					}
				}

				try {
					Thread.sleep(50);
				} catch (InterruptedException e) {
				}

				timeval += 50;
			}
		}

		if (!noBaudRateChange) {
			try {
				serialPort.setSerialPortParams(baudRate, SerialPort.DATABITS_7, SerialPort.STOPBITS_1,
						SerialPort.PARITY_EVEN);
			} catch (UnsupportedCommOperationException e) {
				close();
				throw new IOException("Serial Port does not support " + baudRate + "bd 8N1");
			}

		}

		readSuccessful = false;
		numBytesRead = 0;

		while (timeout == 0 || timeval < timeout) {
			if (is.available() > 0) {
				numBytesRead += is.read(buffer, numBytesRead, INPUT_BUFFER_LENGTH - numBytesRead);

				if ((numBytesRead > 4 && buffer[numBytesRead - 3] == 0x21)
						|| (numBytesRead > 7 && buffer[numBytesRead - 5] == 0x21)) {
					readSuccessful = true;
					break;
				}
			}

			try {
				Thread.sleep(50);
			} catch (InterruptedException e) {
			}

			timeval += 50;
		}

		if (!readSuccessful) {
			return null;
		}

		int index;
		int endIndex;
		if (buffer[numBytesRead - 2] == 0x03) {
			if (numBytesRead < 8) {
				throw new IOException("Data message does not have minimum length of 8.");
			}
			index = 1;
			endIndex = numBytesRead - 5;
			if (buffer[0] != 0x02) {
				throw new IOException(
						"STX (0x02) character is expected but not received as first byte of data message.");
			}
		}
		else {
			if (numBytesRead < 5) {
				throw new IOException("Data message does not have minimum length of 5.");
			}
			index = 0;
			endIndex = numBytesRead - 3;
		}

		if (buffer[endIndex + 1] != 0x0D) {
			throw new IOException("CR (0x0D) character is expected but not received after data block of data message.");
		}

		if (buffer[endIndex + 2] != 0x0A) {
			throw new IOException("LF (0x0A) character is expected but not received after data block of data message.");
		}

		List<DataSet> dataSets = new ArrayList<DataSet>();

		while (index != endIndex) {
			String id = null;

			for (int i = index; i < endIndex - 1; i++) {
				// find '('
				if (buffer[i] == 0x28) {
					id = new String(buffer, index, i - index);
					index = i + 1;
					break;
				}
			}
			if (id == null) {
				throw new IOException(
						"'(' (0x28) character is expected but not received inside data block of data message.");
			}

			String value = "";
			String unit = "";
			for (int i = index; i < endIndex; i++) {
				// found '*'
				if (buffer[i] == 0x2A) {
					if (i > index) {
						value = new String(buffer, index, i - index);
					}
					index = i + 1;

					for (int j = index; j < endIndex - 1; j++) {
						// found ')'
						if (buffer[j] == 0x29) {
							unit = new String(buffer, index, j - index);
							index = j + 1;
							break;
						}
					}

					break;
				}
				// found ')'
				else if (buffer[i] == 0x29) {
					if (i > index) {
						value = new String(buffer, index, i - index);
					}
					index = i + 1;
					break;
				}
			}
			if (buffer[index - 1] != 0x29) {
				throw new IOException(
						"')' (0x29) character is expected but not received inside data block of data message.");
			}

			dataSets.add(new DataSet(id, value, unit));

			if (buffer[index] == 0x0d && buffer[index + 1] == 0x0a) {
				index += 2;
			}

		}
		return dataSets;

	}

	/**
	 * Returns the baud rate chosen by the server for this communication
	 * 
	 * @param baudCharacter
	 *            Encoded baud rate (see 6.3.14 item 13c)
	 * @return The chosen baud rate or -1 on error
	 */
	private int getBaudRate(int baudCharacter) {
		int result = -1;
		switch (baudCharacter) {
		case 0x30:
			result = 300;
			break;
		case 0x31:
			result = 600;
			break;
		case 0x32:
			result = 1200;
			break;
		case 0x33:
			result = 2400;
			break;
		case 0x34:
			result = 4800;
			break;
		case 0x35:
			result = 9600;
			break;
		case 0x36:
			result = 19200;
			break;
		}
		return result;
	}

	public void close() {
		serialPort.removeEventListener();
		serialPort.close();
	}

}
